#include <asm/unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <asm/signal.h>
#include <sys/time.h>
#include <asm/fcntl.h>

pid_t waitpid(pid_t, int *, int);
pid_t fork(void);
int execve(const char *, char *const [], char *const []) __attribute__(( __noreturn__ ));
int sethostname(const char *, size_t);
int setdomainname(const char *, size_t);
int kill(pid_t, int);
int open(const char *, int);
int close(int);
int pause(void);
ssize_t read(int, void *, size_t);
ssize_t write(int, const void *, size_t);
void (*signal(int, void (*)(int)))(int);
void exit(int) __attribute__(( __noreturn__ ));
void chop(char *);
void printf(const char *, ...);

extern char str_autoopt;
extern char str_confpath;
extern char str_term;
extern char str_hostname;
extern char str_domainname;
extern char str_spawn;
extern char bannerfmt;
extern char devnull;
#ifdef VERBOSE
extern char respawnfmt;
extern char termfmt;
#endif /* VERBOSE */

#define MSGLEN		512
#define MAXCOMLEN	512
#define MAXOPTLEN	32
#define MAXARG		8
#define MAXENV		8
#define MAXSPAWNS	8

int run(const char *command, const int w)
{
	char	copy[strlen(command)+1];
	char	*a[MAXARG];
	char	*e[MAXENV];
	pid_t	pid;
	int	i;

	if ((pid = fork()) == 0)
	{
		a[0] = copy;
		a[1] = copy;
		i = 1;
		while(*command)
		{
			if (*command == ' ')
			{
				*(a[i]++) = 0;
				a[i+1] = a[i];
				i++;
				while(*command == ' ')
					command++;
			}
			else
			{
				*(a[i]++) = *(command++);
			}
		}
		a[i] = 0;
		a[i] = NULL;
		e[0] = NULL;
		execve(a[0],(char **)&a[1],(char **)e);
		/* NOT REACHED */
	}
	if (w)
	{
		/*
		 *  at this stage, size is more important
		 */
		waitpid(pid,NULL,0);
	}
	return(pid);
}

void atomic_optget(const char *opt, char *buffer, int bufsize)
{
	char	path[256];
	char	*dest,*src;
	int	n,fd;

	dest = path;
	src = &str_confpath;
	while(*src)
		*(dest++) = *(src++);
	while(*opt)
		*(dest++) = *(opt++);
	*dest = 0;

	if ((fd = open(path,O_RDONLY)) < 0)
	{
		*buffer = 0;
		return;
	}
	n = read(fd,buffer,bufsize-1);
	buffer[n] = 0;
	chop(buffer);	/* remove \n and anything after */
	close(fd);
}

void sigterm(int sig)
{
	close(0);
	signal(SIGTERM,sigterm);
}

void sigchld(int sig)
{
	waitpid(-1,NULL,WNOHANG);
	signal(SIGCHLD,sigchld);
}

int main(int argc, char **argv, char **envp) __attribute__(( __noreturn__ ));
int main(int argc, char **argv, char **envp)
{
	char	*runarg;

	runarg = (argv[1]) ? argv[1] : &str_autoopt;

	/*
	 *  banner their asses!
	 */
	printf(&bannerfmt,runarg);

	/*
	 *  smaller than calling signal(SIGCHLD,sigchld)
	 */
	sigchld(0);
	sigterm(0);			/* also does close(0) */
	open(&devnull,O_RDONLY);	/* returns 0 */

	/*
	 *  sethostname and setdomainname
	 */
	{
		char	name[128];
		char	*p;

		atomic_optget(&str_hostname,name,sizeof(name));
		if (name[0])
		{
			for(p=name;*p;p++)
				;
			sethostname(name,p-name);
		}
		atomic_optget(&str_domainname,name,sizeof(name));
		if (name[0])
		{
			for(p=name;*p;p++)
				;
			setdomainname(name,p-name);
		}
	}

	/*
	 *  figure out what to run
	 */
	{
		char	buf[MAXCOMLEN];

		atomic_optget(runarg,buf,sizeof(buf));
		if (buf[0])
			run(buf,1);
	}
	/*
	 *  major kludge-alert
	 *  and no, this is not an entry for the obfuscated C contest...
	 *  (atleast not intended as such)
	 */
	{
		struct	respawn { pid_t pid; char run[MAXOPTLEN]; } spawn[MAXSPAWNS];
		int	x,ret;

		for(x=0;x<MAXSPAWNS;x++)
			spawn[x].pid = 1;

		/*
		 *  read a list of spawns
		 */
		{
			char	buf[MAXCOMLEN];
			char	*dest,*src;

			atomic_optget(&str_spawn,buf,sizeof(buf));
			if (buf[0])
			{
				x = 0;
				src = buf;
			nextspawn:
				dest = spawn[x].run;
				spawn[x].pid = -2;
				while(*src && *src != ' ')
				{
					*(dest++) = *(src++);
				}
				*dest = 0;
				if (*src == ' ')
				{
					x++;
					while(*src == ' ')
						src++;
					if (*src && (x < MAXSPAWNS))
						goto nextspawn;
				}
			}
		}
		goto first_check;
		/*
		 *  this is where we handle respawns and such...
		 */
loop:
		pause();
		/*
		 *  after each signal (SIGCHLD) we come here
		 */
		/* dummy read */
		ret = read(0,&x,1);
		/*
		 *  stdin is open unless we have gotten a SIGTERM
		 */
		if (ret != 0)
		{
			char	buf[MAXCOMLEN];

			/*
			 *  ignore SIGTERM while we're in the shutdown procedure
			 */
			signal(SIGTERM,SIG_IGN);
			atomic_optget(&str_term,buf,sizeof(buf));
#ifdef VERBOSE
			printf(&termfmt,buf);
#endif /* VERBOSE */
			run(buf,0);
		forever:
			pause();
			goto forever;
		}
first_check:
		for(x=0;x<MAXSPAWNS;x++)
		{
			ret = kill(spawn[x].pid,0);
			if (ret < 0)
			{
				char	buf[MAXCOMLEN];

				atomic_optget(spawn[x].run,buf,sizeof(buf));
				if (buf[0])
				{
					spawn[x].pid = run(buf,0);
#ifdef VERBOSE
					/*
					 *  buf is chopped in child only
					 */
					printf(&respawnfmt,buf,x+1,spawn[x].run,spawn[x].pid);
#endif /* VERBOSE */
				}
			}
		}
		goto loop;
	}
}

